@@ -7,6 +7,7 @@
  */
 #include <stdlib.h>
 #include <erofs/cache.h>
+#include <erofs/bitops.h>
 #include "erofs/print.h"
 
 static int erofs_bh_flush_drop_directly(struct erofs_buffer_head *bh)
@@ -30,38 +31,74 @@ const struct erofs_bhops erofs_skip_write_bhops = {
 struct erofs_bufmgr *erofs_buffer_init(struct erofs_sb_info *sbi,
 				       erofs_blk_t startblk)
 {
-	struct erofs_bufmgr *bufmgr;
-	int i, j;
+	unsigned int blksiz = erofs_blksiz(sbi);
+	struct erofs_bufmgr *bmgr;
+	int i, j, k;
 
-	bufmgr = malloc(sizeof(struct erofs_bufmgr));
-	if (!bufmgr)
+	bmgr = malloc(sizeof(struct erofs_bufmgr));
+	if (!bmgr)
 		return NULL;
 
-	init_list_head(&bufmgr->blkh.list);
-	bufmgr->blkh.blkaddr = NULL_ADDR;
-	bufmgr->last_mapped_block = &bufmgr->blkh;
-
-	for (i = 0; i < ARRAY_SIZE(bufmgr->mapped_buckets); i++)
-		for (j = 0; j < ARRAY_SIZE(bufmgr->mapped_buckets[0]); j++)
-			init_list_head(&bufmgr->mapped_buckets[i][j]);
-	bufmgr->tail_blkaddr = startblk;
-	bufmgr->sbi = sbi;
-	return bufmgr;
+	bmgr->sbi = sbi;
+	for (i = 0; i < ARRAY_SIZE(bmgr->watermeter); i++) {
+		for (j = 0; j < ARRAY_SIZE(bmgr->watermeter[0]); j++) {
+			for (k = 0; k < blksiz; k++)
+				init_list_head(&bmgr->watermeter[i][j][k]);
+			memset(bmgr->bktmap[i][j], 0,
+			       (blksiz / BITS_PER_LONG) * sizeof(unsigned long));
+		}
+	}
+	init_list_head(&bmgr->blkh.list);
+	bmgr->blkh.blkaddr = NULL_ADDR;
+	bmgr->tail_blkaddr = startblk;
+	bmgr->last_mapped_block = &bmgr->blkh;
+	bmgr->dsunit = 0;
+	return bmgr;
 }
 
-static void erofs_bupdate_mapped(struct erofs_buffer_block *bb)
+static void erofs_clear_bbktmap(struct erofs_bufmgr *bmgr, int type,
+				bool mapped, int nr)
+{
+	int bit = erofs_blksiz(bmgr->sbi) - (nr + 1);
+
+	DBG_BUGON(bit < 0);
+	__erofs_clear_bit(bit, bmgr->bktmap[type][mapped]);
+}
+
+static void erofs_set_bbktmap(struct erofs_bufmgr *bmgr, int type,
+			      bool mapped, int nr)
+{
+	int bit = erofs_blksiz(bmgr->sbi) - (nr + 1);
+
+	DBG_BUGON(bit < 0);
+	__erofs_set_bit(bit, bmgr->bktmap[type][mapped]);
+}
+
+static void erofs_update_bwatermeter(struct erofs_buffer_block *bb, bool free)
 {
 	struct erofs_bufmgr *bmgr = bb->buffers.fsprivate;
 	struct erofs_sb_info *sbi = bmgr->sbi;
-	struct list_head *bkt;
+	unsigned int blksiz = erofs_blksiz(sbi);
+	bool mapped = bb->blkaddr != NULL_ADDR;
+	struct list_head *h = bmgr->watermeter[bb->type][mapped];
+	unsigned int nr;
 
-	if (bb->blkaddr == NULL_ADDR)
+	if (bb->sibling.next == bb->sibling.prev) {
+		if ((uintptr_t)(bb->sibling.next - h) < blksiz) {
+			nr = bb->sibling.next - h;
+			erofs_clear_bbktmap(bmgr, bb->type, mapped, nr);
+		} else if (bb->sibling.next != &bb->sibling) {
+			nr = bb->sibling.next -
+				bmgr->watermeter[bb->type][!mapped];
+			erofs_clear_bbktmap(bmgr, bb->type, !mapped, nr);
+		}
+	}
+	list_del(&bb->sibling);
+	if (free)
 		return;
-
-	bkt = bmgr->mapped_buckets[bb->type] +
-		(bb->buffers.off & (erofs_blksiz(sbi) - 1));
-	list_del(&bb->mapped_list);
-	list_add_tail(&bb->mapped_list, bkt);
+	nr = bb->buffers.off & (blksiz - 1);
+	list_add_tail(&bb->sibling, h + nr);
+	erofs_set_bbktmap(bmgr, bb->type, mapped, nr);
 }
 
 /* return occupied bytes in specific buffer block if succeed */
@@ -69,20 +106,26 @@ static int __erofs_battach(struct erofs_buffer_block *bb,
 			   struct erofs_buffer_head *bh,
 			   erofs_off_t incr,
 			   unsigned int alignsize,
-			   unsigned int extrasize,
+			   unsigned int inline_ext,
 			   bool dryrun)
 {
 	struct erofs_bufmgr *bmgr = bb->buffers.fsprivate;
 	struct erofs_sb_info *sbi = bmgr->sbi;
-	const unsigned int blksiz = erofs_blksiz(sbi);
-	const unsigned int blkmask = blksiz - 1;
+	const unsigned int blkmask = erofs_blksiz(sbi) - 1;
 	erofs_off_t boff = bb->buffers.off;
-	const erofs_off_t alignedoffset = roundup(boff, alignsize);
-	const int oob = cmpsgn(roundup(((boff - 1) & blkmask) + 1, alignsize) +
-					incr + extrasize, blksiz);
+	const erofs_off_t alignedoffset = round_up(boff, alignsize);
 	bool tailupdate = false;
 	erofs_blk_t blkaddr;
+	int oob;
 
+	DBG_BUGON(alignsize & (alignsize - 1));
+	/* inline data must never span block boundaries */
+	if (erofs_blkoff(sbi, alignedoffset + incr + blkmask)
+			+ inline_ext > blkmask)
+		return -ENOSPC;
+
+	oob = cmpsgn(alignedoffset + incr + inline_ext,
+		     bb->buffers.nblocks << sbi->blkszbits);
 	if (oob >= 0) {
 		/* the next buffer block should be NULL_ADDR all the time */
 		if (oob && list_next_entry(bb, list)->blkaddr != NULL_ADDR)
@@ -91,7 +134,7 @@ static int __erofs_battach(struct erofs_buffer_block *bb,
 		blkaddr = bb->blkaddr;
 		if (blkaddr != NULL_ADDR) {
 			tailupdate = (bmgr->tail_blkaddr == blkaddr +
-				      BLK_ROUND_UP(sbi, boff));
+				      bb->buffers.nblocks);
 			if (oob && !tailupdate)
 				return -EINVAL;
 		}
@@ -105,13 +148,14 @@ static int __erofs_battach(struct erofs_buffer_block *bb,
 		}
 		boff = alignedoffset + incr;
 		bb->buffers.off = boff;
+		bb->buffers.nblocks = max_t(erofs_blk_t, bb->buffers.nblocks,
+					    BLK_ROUND_UP(sbi, boff));
 		/* need to update the tail_blkaddr */
 		if (tailupdate)
-			bmgr->tail_blkaddr = blkaddr +
-						BLK_ROUND_UP(sbi, boff);
-		erofs_bupdate_mapped(bb);
+			bmgr->tail_blkaddr = blkaddr + bb->buffers.nblocks;
+		erofs_update_bwatermeter(bb, false);
 	}
-	return ((alignedoffset + incr - 1) & blkmask) + 1;
+	return ((alignedoffset + incr + blkmask) & blkmask) + 1;
 }
 
 int erofs_bh_balloon(struct erofs_buffer_head *bh, erofs_off_t incr)
@@ -125,108 +169,125 @@ int erofs_bh_balloon(struct erofs_buffer_head *bh, erofs_off_t incr)
 	return __erofs_battach(bb, NULL, incr, 1, 0, false);
 }
 
+static bool __find_next_bucket(struct erofs_bufmgr *bmgr, int type, bool mapped,
+			       unsigned int *index, unsigned int end)
+{
+	const unsigned int blksiz = erofs_blksiz(bmgr->sbi);
+	const unsigned int blkmask = blksiz - 1;
+	unsigned int l = *index, r;
+
+	if (l <= end) {
+		DBG_BUGON(l < end);
+		return false;
+	}
+
+	l = blkmask - (l & blkmask);
+	r = blkmask - (end & blkmask);
+	if (l >= r) {
+		l = erofs_find_next_bit(bmgr->bktmap[type][mapped], blksiz, l);
+		if (l < blksiz) {
+			*index = round_down(*index, blksiz) + blkmask - l;
+			return true;
+		}
+		l = 0;
+		*index -= blksiz;
+	}
+	l = erofs_find_next_bit(bmgr->bktmap[type][mapped], r, l);
+	if (l >= r)
+		return false;
+	*index = round_down(*index, blksiz) + blkmask - l;
+	return true;
+}
+
 static int erofs_bfind_for_attach(struct erofs_bufmgr *bmgr,
 				  int type, erofs_off_t size,
-				  unsigned int required_ext,
 				  unsigned int inline_ext,
 				  unsigned int alignsize,
 				  struct erofs_buffer_block **bbp)
 {
 	const unsigned int blksiz = erofs_blksiz(bmgr->sbi);
+	const unsigned int blkmask = blksiz - 1;
 	struct erofs_buffer_block *cur, *bb;
-	unsigned int used0, used_before, usedmax, used;
+	unsigned int index, used0, end, mapped;
+	unsigned int usedmax, used;
 	int ret;
 
-	used0 = ((size + required_ext) & (blksiz - 1)) + inline_ext;
-	/* inline data should be in the same fs block */
-	if (used0 > blksiz)
-		return -ENOSPC;
-
-	if (!used0 || alignsize == blksiz) {
+	if (alignsize == blksiz) {
 		*bbp = NULL;
 		return 0;
 	}
-
 	usedmax = 0;
 	bb = NULL;
 
-	/* try to find a most-fit mapped buffer block first */
-	if (size + required_ext + inline_ext >= blksiz)
-		goto skip_mapped;
-
-	used_before = rounddown(blksiz -
-				(size + required_ext + inline_ext), alignsize);
-	for (; used_before; --used_before) {
-		struct list_head *bt = bmgr->mapped_buckets[type] + used_before;
-
-		if (list_empty(bt))
-			continue;
-		cur = list_first_entry(bt, struct erofs_buffer_block,
-				       mapped_list);
-
-		/* last mapped block can be expended, don't handle it here */
-		if (list_next_entry(cur, list)->blkaddr == NULL_ADDR) {
-			DBG_BUGON(cur != bmgr->last_mapped_block);
-			continue;
-		}
-
-		DBG_BUGON(cur->type != type);
-		DBG_BUGON(cur->blkaddr == NULL_ADDR);
-		DBG_BUGON(used_before != (cur->buffers.off & (blksiz - 1)));
-
-		ret = __erofs_battach(cur, NULL, size, alignsize,
-				      required_ext + inline_ext, true);
-		if (ret < 0) {
-			DBG_BUGON(1);
-			continue;
-		}
-
-		/* should contain all data in the current block */
-		used = ret + required_ext + inline_ext;
-		DBG_BUGON(used > blksiz);
-
-		bb = cur;
-		usedmax = used;
-		break;
+	mapped = ARRAY_SIZE(bmgr->watermeter);
+	used0 = rounddown(blksiz - ((size + inline_ext) & blkmask), alignsize);
+	if (__erofs_unlikely(bmgr->dsunit > 1)) {
+		end = used0 + alignsize - 1;
+	} else {
+		end = blksiz;
+		if (size + inline_ext >= blksiz)
+			--mapped;
 	}
+	index = used0 + blksiz;
 
-skip_mapped:
-	/* try to start from the last mapped one, which can be expended */
-	cur = bmgr->last_mapped_block;
-	if (cur == &bmgr->blkh)
-		cur = list_next_entry(cur, list);
-	for (; cur != &bmgr->blkh; cur = list_next_entry(cur, list)) {
-		used_before = cur->buffers.off & (blksiz - 1);
+	while (mapped) {
+		--mapped;
+		for (; __find_next_bucket(bmgr, type, mapped, &index, end);
+		     --index) {
+			struct list_head *bt;
 
-		/* skip if buffer block is just full */
-		if (!used_before)
-			continue;
+			used = index & blkmask;
+			bt = bmgr->watermeter[type][mapped] + used;
+			DBG_BUGON(list_empty(bt));
+			cur = list_first_entry(bt, struct erofs_buffer_block,
+					       sibling);
 
-		/* skip if the entry which has different type */
-		if (cur->type != type)
-			continue;
+			/* skip the last mapped block */
+			if (mapped &&
+			    list_next_entry(cur, list)->blkaddr == NULL_ADDR) {
+				DBG_BUGON(cur != bmgr->last_mapped_block);
+				cur = list_next_entry(cur, sibling);
+				if (&cur->sibling == bt)
+					continue;
+			}
 
-		ret = __erofs_battach(cur, NULL, size, alignsize,
-				      required_ext + inline_ext, true);
-		if (ret < 0)
-			continue;
+			DBG_BUGON(cur->type != type);
+			DBG_BUGON((cur->blkaddr != NULL_ADDR) ^ mapped);
+			DBG_BUGON(used != (cur->buffers.off & blkmask));
 
-		used = ((ret + required_ext) & (blksiz - 1)) + inline_ext;
+			ret = __erofs_battach(cur, NULL, size, alignsize,
+					      inline_ext, true);
+			if (ret < 0) {
+				DBG_BUGON(mapped && !(bmgr->dsunit > 1));
+				continue;
+			}
 
-		/* should contain inline data in current block */
-		if (used > blksiz)
-			continue;
+			used = ret + inline_ext;
 
-		/*
-		 * remaining should be smaller than before or
-		 * larger than allocating a new buffer block
-		 */
-		if (used < used_before && used < used0)
-			continue;
+			/* should contain all data in the current block */
+			DBG_BUGON(used > blksiz);
+			if (used > usedmax) {
+				usedmax = used;
+				bb = cur;
+			}
+			break;
+		}
+		end = used0 + alignsize - 1;
+		index = used0 + blksiz;
 
-		if (usedmax < used) {
-			bb = cur;
-			usedmax = used;
+		/* try the last mapped block independently */
+		cur = bmgr->last_mapped_block;
+		if (mapped && cur != &bmgr->blkh && cur->type == type) {
+			ret = __erofs_battach(cur, NULL, size,
+					      alignsize, inline_ext, true);
+			if (ret >= 0) {
+				used = ret + inline_ext;
+				DBG_BUGON(used > blksiz);
+				if (used > usedmax) {
+					usedmax = used;
+					bb = cur;
+				}
+			}
 		}
 	}
 	*bbp = bb;
@@ -235,7 +296,6 @@ skip_mapped:
 
 struct erofs_buffer_head *erofs_balloc(struct erofs_bufmgr *bmgr,
 				       int type, erofs_off_t size,
-				       unsigned int required_ext,
 				       unsigned int inline_ext)
 {
 	struct erofs_buffer_block *bb;
@@ -251,7 +311,7 @@ struct erofs_buffer_head *erofs_balloc(struct erofs_bufmgr *bmgr,
 	alignsize = ret;
 
 	/* try to find if we could reuse an allocated buffer block */
-	ret = erofs_bfind_for_attach(bmgr, type, size, required_ext, inline_ext,
+	ret = erofs_bfind_for_attach(bmgr, type, size, inline_ext,
 				     alignsize, &bb);
 	if (ret)
 		return ERR_PTR(ret);
@@ -269,6 +329,7 @@ struct erofs_buffer_head *erofs_balloc(struct erofs_bufmgr *bmgr,
 		bb->type = type;
 		bb->blkaddr = NULL_ADDR;
 		bb->buffers.off = 0;
+		bb->buffers.nblocks = 0;
 		bb->buffers.fsprivate = bmgr;
 		init_list_head(&bb->buffers.list);
 		if (type == DATA)
@@ -276,7 +337,7 @@ struct erofs_buffer_head *erofs_balloc(struct erofs_bufmgr *bmgr,
 				 &bmgr->last_mapped_block->list);
 		else
 			list_add_tail(&bb->list, &bmgr->blkh.list);
-		init_list_head(&bb->mapped_list);
+		init_list_head(&bb->sibling);
 
 		bh = malloc(sizeof(struct erofs_buffer_head));
 		if (!bh) {
@@ -285,8 +346,7 @@ struct erofs_buffer_head *erofs_balloc(struct erofs_bufmgr *bmgr,
 		}
 	}
 
-	ret = __erofs_battach(bb, bh, size, alignsize,
-			      required_ext + inline_ext, false);
+	ret = __erofs_battach(bb, bh, size, alignsize, inline_ext, false);
 	if (ret < 0) {
 		free(bh);
 		return ERR_PTR(ret);
@@ -323,21 +383,30 @@ struct erofs_buffer_head *erofs_battach(struct erofs_buffer_head *bh,
 	return nbh;
 }
 
-static erofs_blk_t __erofs_mapbh(struct erofs_buffer_block *bb)
+static void __erofs_mapbh(struct erofs_buffer_block *bb)
 {
 	struct erofs_bufmgr *bmgr = bb->buffers.fsprivate;
-	erofs_blk_t blkaddr;
+	erofs_blk_t blkaddr = bmgr->tail_blkaddr;
 
 	if (bb->blkaddr == NULL_ADDR) {
-		bb->blkaddr = bmgr->tail_blkaddr;
+		bb->blkaddr = blkaddr;
+		if (__erofs_unlikely(bmgr->dsunit > 1) && bb->type == DATA) {
+			struct erofs_buffer_block *pb = list_prev_entry(bb, list);
+
+			bb->blkaddr = roundup(blkaddr, bmgr->dsunit);
+			if (pb != &bmgr->blkh &&
+			    pb->blkaddr + pb->buffers.nblocks >= blkaddr) {
+				DBG_BUGON(pb->blkaddr + pb->buffers.nblocks > blkaddr);
+				pb->buffers.nblocks = bb->blkaddr - pb->blkaddr;
+			}
+		}
 		bmgr->last_mapped_block = bb;
-		erofs_bupdate_mapped(bb);
+		erofs_update_bwatermeter(bb, false);
 	}
 
-	blkaddr = bb->blkaddr + BLK_ROUND_UP(bmgr->sbi, bb->buffers.off);
+	blkaddr = bb->blkaddr + bb->buffers.nblocks;
 	if (blkaddr > bmgr->tail_blkaddr)
 		bmgr->tail_blkaddr = blkaddr;
-	return blkaddr;
 }
 
 erofs_blk_t erofs_mapbh(struct erofs_bufmgr *bmgr,
@@ -357,7 +426,7 @@ erofs_blk_t erofs_mapbh(struct erofs_bufmgr *bmgr,
 			break;
 
 		DBG_BUGON(t->blkaddr != NULL_ADDR);
-		(void)__erofs_mapbh(t);
+		__erofs_mapbh(t);
 	} while (t != bb);
 	return bmgr->tail_blkaddr;
 }
@@ -371,7 +440,7 @@ static void erofs_bfree(struct erofs_buffer_block *bb)
 	if (bb == bmgr->last_mapped_block)
 		bmgr->last_mapped_block = list_prev_entry(bb, list);
 
-	list_del(&bb->mapped_list);
+	erofs_update_bwatermeter(bb, true);
 	list_del(&bb->list);
 	free(bb);
 }
@@ -393,7 +462,8 @@ int erofs_bflush(struct erofs_bufmgr *bmgr,
 		if (p == bb)
 			break;
 
-		blkaddr = __erofs_mapbh(p);
+		__erofs_mapbh(p);
+		blkaddr = p->blkaddr + BLK_ROUND_UP(sbi, p->buffers.off);
 
 		list_for_each_entry_safe(bh, nbh, &p->buffers.list, list) {
 			if (bh->op == &erofs_skip_write_bhops) {
@@ -416,8 +486,7 @@ int erofs_bflush(struct erofs_bufmgr *bmgr,
 					   padding, true);
 
 		if (p->type != DATA)
-			bmgr->metablkcnt +=
-				BLK_ROUND_UP(sbi, p->buffers.off);
+			bmgr->metablkcnt += p->buffers.nblocks;
 		erofs_dbg("block %u to %u flushed", p->blkaddr, blkaddr - 1);
 		erofs_bfree(p);
 	}
@@ -428,13 +497,12 @@ void erofs_bdrop(struct erofs_buffer_head *bh, bool tryrevoke)
 {
 	struct erofs_buffer_block *const bb = bh->block;
 	struct erofs_bufmgr *bmgr = bb->buffers.fsprivate;
-	struct erofs_sb_info *sbi = bmgr->sbi;
 	const erofs_blk_t blkaddr = bh->block->blkaddr;
 	bool rollback = false;
 
 	/* tail_blkaddr could be rolled back after revoking all bhs */
 	if (tryrevoke && blkaddr != NULL_ADDR &&
-	    bmgr->tail_blkaddr == blkaddr + BLK_ROUND_UP(sbi, bb->buffers.off))
+	    bmgr->tail_blkaddr == blkaddr + bb->buffers.nblocks)
 		rollback = true;
 
 	bh->op = &erofs_drop_directly_bhops;
@@ -444,7 +512,7 @@ void erofs_bdrop(struct erofs_buffer_head *bh, bool tryrevoke)
 		return;
 
 	if (!rollback && bb->type != DATA)
-		bmgr->metablkcnt += BLK_ROUND_UP(sbi, bb->buffers.off);
+		bmgr->metablkcnt += bb->buffers.nblocks;
 	erofs_bfree(bb);
 	if (rollback)
 		bmgr->tail_blkaddr = blkaddr;
