@@ -3,9 +3,6 @@
  * Copyright (C), 2022, Coolpad Group Limited.
  * Created by Yue Hu <huyue2@coolpad.com>
  */
-#ifndef _LARGEFILE_SOURCE
-#define _LARGEFILE_SOURCE
-#endif
 #ifndef _LARGEFILE64_SOURCE
 #define _LARGEFILE64_SOURCE
 #endif
@@ -25,6 +22,7 @@
 #include "erofs/internal.h"
 #include "erofs/fragments.h"
 #include "erofs/bitops.h"
+#include "liberofs_private.h"
 
 struct erofs_fragment_dedupe_item {
 	struct list_head	list;
@@ -33,7 +31,7 @@ struct erofs_fragment_dedupe_item {
 	u8			data[];
 };
 
-#define EROFS_FRAGMENT_INMEM_SZ_MAX	EROFS_CONFIG_COMPR_MAX_SZ
+#define EROFS_FRAGMENT_INMEM_SZ_MAX	(256 * 1024)
 #define EROFS_TOF_HASHLEN		16
 
 #define FRAGMENT_HASHSIZE		65536
@@ -41,12 +39,12 @@ struct erofs_fragment_dedupe_item {
 
 struct erofs_packed_inode {
 	struct list_head *hash;
-	FILE *file;
+	int fd;
 	unsigned long *uptodate;
 #if EROFS_MT_ENABLED
 	pthread_mutex_t mutex;
 #endif
-	unsigned int uptodate_size;
+	u64 uptodate_bits;
 };
 
 const char *erofs_frags_packedname = "packed_file";
@@ -108,8 +106,7 @@ static int z_erofs_fragments_dedupe_find(struct erofs_inode *inode, int fd,
 
 				sz = min_t(u64, pos, sizeof(buf[0]));
 				sz = min_t(u64, sz, inode->i_size - i);
-				if (pread(fileno(epi->file), buf[0], sz,
-					  pos - sz) != sz)
+				if (pread(epi->fd, buf[0], sz, pos - sz) != sz)
 					break;
 				if (pread(fd, buf[1], sz,
 					  inode->i_size - i - sz) != sz)
@@ -149,21 +146,13 @@ int z_erofs_fragments_dedupe(struct erofs_inode *inode, int fd, u32 *tofcrc)
 	if (inode->i_size <= EROFS_TOF_HASHLEN)
 		return 0;
 
-	if (erofs_lseek64(fd, inode->i_size - EROFS_TOF_HASHLEN, SEEK_SET) < 0)
-		return -errno;
-
-	ret = read(fd, data_to_hash, EROFS_TOF_HASHLEN);
+	ret = pread(fd, data_to_hash, EROFS_TOF_HASHLEN,
+		    inode->i_size - EROFS_TOF_HASHLEN);
 	if (ret != EROFS_TOF_HASHLEN)
 		return -errno;
 
 	*tofcrc = erofs_crc32c(~0, data_to_hash, EROFS_TOF_HASHLEN);
-	ret = z_erofs_fragments_dedupe_find(inode, fd, *tofcrc);
-	if (ret < 0)
-		return ret;
-	ret = lseek(fd, 0, SEEK_SET);
-	if (ret < 0)
-		return -errno;
-	return 0;
+	return z_erofs_fragments_dedupe_find(inode, fd, *tofcrc);
 }
 
 static int z_erofs_fragments_dedupe_insert(struct list_head *hash, void *data,
@@ -190,32 +179,13 @@ static int z_erofs_fragments_dedupe_insert(struct list_head *hash, void *data,
 	return 0;
 }
 
-void z_erofs_fragments_commit(struct erofs_inode *inode)
-{
-	if (!inode->fragment_size)
-		return;
-	/*
-	 * If the packed inode is larger than 4GiB, the full fragmentoff
-	 * will be recorded by switching to the noncompact layout anyway.
-	 */
-	if (inode->fragmentoff >> 32)
-		inode->datalayout = EROFS_INODE_COMPRESSED_FULL;
-
-	inode->z_advise |= Z_EROFS_ADVISE_FRAGMENT_PCLUSTER;
-	erofs_sb_set_fragments(inode->sbi);
-}
-
 int z_erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofcrc)
 {
 	struct erofs_packed_inode *epi = inode->sbi->packedinode;
-#ifdef HAVE_FTELLO64
-	off64_t offset = ftello64(epi->file);
-#else
-	off_t offset = ftello(epi->file);
-#endif
+	s64 offset, rc;
 	char *memblock;
-	int rc;
 
+	offset = lseek(epi->fd, 0, SEEK_CUR);
 	if (offset < 0)
 		return -errno;
 
@@ -234,14 +204,19 @@ int z_erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofcrc)
 
 			rc = read(fd, buf, sz);
 			if (rc != sz) {
-				if (rc < 0)
-					rc = -errno;
-				else
-					rc = -EAGAIN;
-				goto out;
+				if (rc <= 0) {
+					if (!rc)
+						rc = -EIO;
+					else
+						rc = -errno;
+					goto out;
+				}
+				sz = rc;
 			}
-			if (fwrite(buf, sz, 1, epi->file) != 1) {
-				rc = -EIO;
+			rc = __erofs_io_write(epi->fd, buf, sz);
+			if (rc != sz) {
+				if (rc >= 0)
+					rc = -EIO;
 				goto out;
 			}
 			remaining -= sz;
@@ -251,13 +226,18 @@ int z_erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofcrc)
 			rc = -errno;
 			goto out;
 		}
-	} else if (fwrite(memblock, inode->fragment_size, 1, epi->file) != 1) {
-		rc = -EIO;
-		goto out;
+	} else {
+		rc = __erofs_io_write(epi->fd, memblock, inode->fragment_size);
+		if (rc != inode->fragment_size) {
+			if (rc >= 0)
+				rc = -EIO;
+			goto out;
+		}
 	}
 
-	erofs_dbg("Recording %llu fragment data at %llu",
-		  inode->fragment_size | 0ULL, inode->fragmentoff | 0ULL);
+	erofs_dbg("Recording %llu fragment data at %llu of %s",
+		  inode->fragment_size | 0ULL, inode->fragmentoff | 0ULL,
+		  inode->i_srcpath);
 
 	if (memblock)
 		rc = z_erofs_fragments_dedupe_insert(
@@ -266,6 +246,10 @@ int z_erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofcrc)
 	else
 		rc = 0;
 out:
+	if (rc)
+		erofs_err("Failed to record %llu-byte fragment data @ %llu for nid %llu: %d",
+			  inode->fragment_size | 0ULL,
+			  inode->fragmentoff | 0ULL, inode->nid | 0ULL, (int)rc);
 	if (memblock)
 		munmap(memblock, inode->i_size);
 	return rc;
@@ -275,11 +259,7 @@ int z_erofs_pack_fragments(struct erofs_inode *inode, void *data,
 			   unsigned int len, u32 tofcrc)
 {
 	struct erofs_packed_inode *epi = inode->sbi->packedinode;
-#ifdef HAVE_FTELLO64
-	off64_t offset = ftello64(epi->file);
-#else
-	off_t offset = ftello(epi->file);
-#endif
+	s64 offset = lseek(epi->fd, 0, SEEK_CUR);
 	int ret;
 
 	if (offset < 0)
@@ -288,11 +268,16 @@ int z_erofs_pack_fragments(struct erofs_inode *inode, void *data,
 	inode->fragmentoff = (erofs_off_t)offset;
 	inode->fragment_size = len;
 
-	if (fwrite(data, len, 1, epi->file) != 1)
+	ret = write(epi->fd, data, len);
+	if (ret != len) {
+		if (ret < 0)
+			return -errno;
 		return -EIO;
+	}
 
-	erofs_dbg("Recording %llu fragment data at %llu",
-		  inode->fragment_size | 0ULL, inode->fragmentoff | 0ULL);
+	erofs_dbg("Recording %llu fragment data at %llu of %s",
+		  inode->fragment_size | 0ULL, inode->fragmentoff | 0ULL,
+		  inode->i_srcpath);
 
 	ret = z_erofs_fragments_dedupe_insert(&epi->hash[FRAGMENT_HASH(tofcrc)],
 					      data, len, inode->fragmentoff);
@@ -309,19 +294,18 @@ int erofs_flush_packed_inode(struct erofs_sb_info *sbi)
 	if (!epi || !erofs_sb_has_fragments(sbi))
 		return -EINVAL;
 
-	fflush(epi->file);
-	if (!ftello(epi->file))
+	if (lseek(epi->fd, 0, SEEK_CUR) <= 0)
 		return 0;
-	inode = erofs_mkfs_build_special_from_fd(sbi, fileno(epi->file),
+	inode = erofs_mkfs_build_special_from_fd(sbi, epi->fd,
 						 EROFS_PACKED_INODE);
 	sbi->packed_nid = erofs_lookupnid(inode);
 	erofs_iput(inode);
 	return 0;
 }
 
-FILE *erofs_packedfile(struct erofs_sb_info *sbi)
+int erofs_packedfile(struct erofs_sb_info *sbi)
 {
-	return sbi->packedinode->file;
+	return sbi->packedinode->fd;
 }
 
 void erofs_packedfile_exit(struct erofs_sb_info *sbi)
@@ -343,8 +327,8 @@ void erofs_packedfile_exit(struct erofs_sb_info *sbi)
 		free(epi->hash);
 	}
 
-	if (epi->file)
-		fclose(epi->file);
+	if (epi->fd >= 0)
+		close(epi->fd);
 	free(epi);
 	sbi->packedinode = NULL;
 }
@@ -372,14 +356,9 @@ int erofs_packedfile_init(struct erofs_sb_info *sbi, bool fragments_mkfs)
 			init_list_head(&epi->hash[i]);
 	}
 
-	epi->file =
-#ifdef HAVE_TMPFILE64
-		tmpfile64();
-#else
-		tmpfile();
-#endif
-	if (!epi->file) {
-		err = -errno;
+	epi->fd = erofs_tmpfile();
+	if (epi->fd < 0) {
+		err = epi->fd;
 		goto err_out;
 	}
 
@@ -388,6 +367,7 @@ int erofs_packedfile_init(struct erofs_sb_info *sbi, bool fragments_mkfs)
 			.sbi = sbi,
 			.nid = sbi->packed_nid,
 		};
+		s64 offset;
 
 		err = erofs_read_inode_from_disk(&ei);
 		if (err) {
@@ -396,13 +376,14 @@ int erofs_packedfile_init(struct erofs_sb_info *sbi, bool fragments_mkfs)
 			goto err_out;
 		}
 
-		err = fseek(epi->file, ei.i_size, SEEK_SET);
-		if (err) {
+		offset = lseek(epi->fd, ei.i_size, SEEK_SET);
+		if (offset < 0) {
 			err = -errno;
 			goto err_out;
 		}
-		epi->uptodate_size = BLK_ROUND_UP(sbi, ei.i_size) / 8;
-		epi->uptodate = calloc(1, epi->uptodate_size);
+		epi->uptodate_bits = round_up(BLK_ROUND_UP(sbi, ei.i_size),
+					      sizeof(epi->uptodate) * 8);
+		epi->uptodate = calloc(1, epi->uptodate_bits >> 3);
 		if (!epi->uptodate) {
 			err = -ENOMEM;
 			goto err_out;
@@ -468,13 +449,12 @@ static void *erofs_packedfile_preload(struct erofs_inode *pi,
 	if (err)
 		goto err_out;
 
-	fflush(epi->file);
-	err = pwrite(fileno(epi->file), buffer, map->m_llen, map->m_la);
+	err = pwrite(epi->fd, buffer, map->m_llen, map->m_la);
 	if (err < 0) {
 		err = -errno;
 		if (err == -ENOSPC) {
-			memset(epi->uptodate, 0, epi->uptodate_size);
-			(void)!ftruncate(fileno(epi->file), 0);
+			memset(epi->uptodate, 0, epi->uptodate_bits >> 3);
+			(void)!ftruncate(epi->fd, 0);
 		}
 		goto err_out;
 	}
@@ -524,6 +504,11 @@ int erofs_packedfile_read(struct erofs_sb_info *sbi,
 			erofs_blk_t bnr = erofs_blknr(sbi, pos);
 			bool uptodate;
 
+			if (__erofs_unlikely(bnr >= epi->uptodate_bits)) {
+				erofs_err("packed inode EOF exceeded @ %llu",
+					  pos | 0ULL);
+				return -EFSCORRUPTED;
+			}
 			map.m_la = round_down(pos, bsz);
 			len = min_t(erofs_off_t, bsz - (pos & (bsz - 1)),
 				    end - pos);
@@ -553,7 +538,7 @@ int erofs_packedfile_read(struct erofs_sb_info *sbi,
 			if (!uptodate)
 				continue;
 
-			err = pread(fileno(epi->file), buf, len, pos);
+			err = pread(epi->fd, buf, len, pos);
 			if (err < 0)
 				break;
 			if (err == len) {
